import json
import os.path
import os
import re
import shutil
import sqlite3
import subprocess
from typing import Any, Optional, List, Dict

TEST_DIR = os.path.dirname(__file__)
OUTPUT_DIR = os.path.join(TEST_DIR, "output")


def test_default():
    _run_test("default", [])


def test_julian():
    _run_test("julian", ["--date", "julian", "--datetime", "julian"])


def test_posix():
    _run_test("posix", ["--date", "posix", "--datetime", "posix"])


def test_seconds_from_midnight():
    _run_test("seconds", ["--time", "seconds"])


def _run_test(name, args):
    init_output(name)
    dbfile = output_db_file(name)
    for sasfile in sas_fixtures(f"fixtures/{name}"):
        run_command([sasfile, dbfile] + args)

    conn = sqlite3.connect(dbfile)
    conn.row_factory = sqlite3.Row

    for exp in expected_fixtures(f"fixtures/{name}"):
        assert_rows(
            conn,
            table=str(exp["table"]),
            exp_rows=list(exp["rows"]),
            desc=exp.get("desc"),
        )

        assert_col_types(
            conn,
            table=str(exp["table"]),
            exp_cols=list(exp["cols"]),
            desc=exp.get("desc"),
        )


def init_output(dir):
    output_dir = os.path.join(OUTPUT_DIR, dir)
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir)


def output_db_file(dir, name="test.db"):
    return os.path.join(OUTPUT_DIR, dir, name)


def sas_fixtures(dir="fixtures"):
    for fname in os.listdir(os.path.join(TEST_DIR, dir)):
        _, ext = os.path.splitext(fname)
        if ext.lower() == ".sas7bdat":
            yield os.path.join(TEST_DIR, dir, fname)


def expected_fixtures(dir="fixtures"):
    for fname in os.listdir(os.path.join(TEST_DIR, dir)):
        _, ext = os.path.splitext(fname)
        if ext.lower() == ".json":
            exp = {}
            with open(os.path.join(TEST_DIR, dir, fname), "r") as f:
                exp = json.load(f)
            yield exp


def run_command(args):
    subprocess.run(["sas2sqlite"] + args, check=True)


def assert_rows(
    conn: sqlite3.Connection,
    table: str,
    exp_rows: List[Any],
    desc: Optional[str] = None,
):
    if desc is None:
        desc = table
    c = conn.execute(f"SELECT * FROM `{table}` ORDER BY ROWID;")
    act_rows = c.fetchall()
    assert len(act_rows) == len(exp_rows)
    for (i, (act, exp)) in enumerate(zip(act_rows, exp_rows)):
        for k in exp:
            a = act[k]
            e = exp[k]
            assert (
                a == e
            ), f"row {i+1}, column {k}: expected {repr(e)}, was {repr(a)} ({desc})"


def assert_col_types(
    conn: sqlite3.Connection,
    table: str,
    exp_cols: List[Dict[str, str]],
    desc: Optional[str] = None,
):
    if desc is None:
        desc = table

    c = conn.execute(
        "SELECT sql FROM sqlite_master WHERE LOWER(`name`) = ? LIMIT 1",
        (table.lower(),),
    )
    r = c.fetchone()
    assert r is not None, f"Did not find table '{table}'"
    sql = r[0]
    for (i, exp) in enumerate(exp_cols):
        for k in exp:
            exp_type = exp[k]
            assert matches_col_type(
                k, exp_type, sql
            ), f"row {i+1}, column {k}: expected '{exp_type}' ({desc})"


def matches_col_type(col_name, exp_type, sql):
    return re.search(f"`{col_name}` {exp_type}", sql) is not None



from argparse import ArgumentParser
import sqlite3
import sys
from typing import Tuple

from sas7bdat import SAS7BDAT  # type: ignore
from . import import_dataset
from . import adapters


def main(argv=sys.argv):
    cmd = ArgumentParser(description="Import sas7bdat files to sqlite3 dbase")
    cmd.add_argument("sas_file", help="source SAS file")
    cmd.add_argument("db_file", help="target sqlite3 database file")
    cmd.add_argument("-t", "--table", help="table name (defaults to SAS dataset name)")
    cmd.add_argument("--create-table", action="store_true", help="create table")
    cmd.add_argument(
        "--no-create-table",
        action="store_false",
        dest="create_table",
        help="do not create table",
    )
    cmd.add_argument(
        "--time",
        type=parse_time_args,
        default="%H:%M:%S",
        help='store time as "seconds", or strftime pattern',
    )
    cmd.add_argument(
        "--date",
        type=parse_date_args,
        default="%Y-%m-%d",
        help='store date as "julian", "posix", or strftime pattern',
    )
    cmd.add_argument(
        "--datetime",
        type=parse_datetime_args,
        default="%Y-%m-%dT%H:%M:%S",
        help='store datetime as "julian", "posix", or strftime pattern',
    )

    cmd.set_defaults(create_table=True)

    args = cmd.parse_args(argv[1:])

    adapters.store_time(*args.time)
    adapters.store_date(*args.date)
    adapters.store_datetime(*args.datetime)

    conn = sqlite3.connect(args.db_file)
    dataset = SAS7BDAT(args.sas_file, skip_header=True)

    import_dataset(
        conn,
        dataset,
        create_table=args.create_table,
        sqlite_time_type=sqlite_time_type(args.time[0]),
        sqlite_date_type=sqlite_date_type(args.date[0]),
        sqlite_datetime_type=sqlite_datetime_type(args.datetime[0]),
    )


def parse_time_args(s: str) -> Tuple[str, str]:
    if s == "seconds":
        return ("seconds", "")
    else:
        return ("text", s)


def parse_date_args(s: str) -> Tuple[str, str]:
    if s == "julian":
        return ("julian", "")
    elif s == "posix":
        return ("posix", "")
    else:
        return ("text", s)


def parse_datetime_args(s: str) -> Tuple[str, str]:
    if s == "julian":
        return ("julian", "")
    elif s == "posix":
        return ("posix", "")
    else:
        return ("text", s)


def sqlite_time_type(s: str) -> str:
    if s == "seconds":
        return "NUMERIC"
    if s == "text":
        return "TEXT"
    raise ValueError(f"No sqlite time adapter found for '{s}'")


def sqlite_date_type(s: str) -> str:
    if s == "julian":
        return "NUMERIC"
    if s == "posix":
        return "NUMERIC"
    if s == "text":
        return "TEXT"
    raise ValueError(f"No sqlite date adapter found for '{s}'")


def sqlite_datetime_type(s: str) -> str:
    if s == "julian":
        return "NUMERIC"
    if s == "posix":
        return "NUMERIC"
    if s == "text":
        return "TEXT"
    raise ValueError(f"No sqlite datetime adapter found for '{s}'")


if __name__ == "__main__":
    main()



from calendar import timegm
from datetime import date, datetime, time
import sqlite3
from typing import Callable

import julian  # type: ignore


def store_time(time_type: str, time_format: str = "") -> None:
    if time_type == "seconds":
        sqlite3.register_adapter(time, time_to_seconds)
    elif time_type == "text":
        sqlite3.register_adapter(time, time_to_text(time_format))
    else:
        raise ValueError(f"Unknown time adapter: '{time_type}'")


def store_date(date_type: str, date_format: str = "") -> None:
    if date_type == "julian":
        sqlite3.register_adapter(date, date_to_julian)
    elif date_type == "posix":
        sqlite3.register_adapter(date, date_to_posix)
    elif date_type == "text":
        sqlite3.register_adapter(date, date_to_text(date_format))
    else:
        raise ValueError(f"Unknown date adapter: '{date_type}'")


def store_datetime(datetime_type: str, datetime_format: str = "") -> None:
    if datetime_type == "julian":
        sqlite3.register_adapter(datetime, datetime_to_julian)
    elif datetime_type == "posix":
        sqlite3.register_adapter(datetime, datetime_to_posix)
    elif datetime_type == "text":
        sqlite3.register_adapter(datetime, datetime_to_text(datetime_format))
    else:
        raise ValueError(f"Unknown datetime adapter: '{datetime_type}'")


def time_to_seconds(t: time) -> float:
    return (60 * 60 * t.hour) + (60 * t.minute) + t.second + t.microsecond


def time_to_text(format: str) -> Callable[[time], str]:
    def _time_to_text(t: time) -> str:
        return t.strftime(format)

    return _time_to_text


def date_to_posix(d: date) -> int:
    return datetime_to_posix(datetime(d.year, d.month, d.day))


def date_to_julian(d: date) -> float:
    return datetime_to_julian(datetime(d.year, d.month, d.day))


def date_to_text(format: str) -> Callable[[date], str]:
    def _date_to_text(d: date) -> str:
        return d.strftime(format)

    return _date_to_text


def datetime_to_posix(dt: datetime) -> int:
    return timegm(dt.utctimetuple())


def datetime_to_julian(dt: datetime) -> float:
    return float(julian.to_jd(dt))


def datetime_to_text(format: str) -> Callable[[datetime], str]:
    def _datetime_to_text(dt: datetime) -> str:
        return dt.strftime(format)

    return _datetime_to_text



import codecs
import os.path
from setuptools import setup, find_packages

# Note: copied from pip setup.py


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


install_requires = ["sas7bdat", "julian"]

tests_require = ["pytest"]


setup(
    name="sas2sqlite3",
    version=get_version("src/sas2sqlite3/__init__.py"),
    author="Eric Gjertsen",
    author_email="ericgj72@gmail.com",
    description="Import sas7bdat files to sqlite3 dbase",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/ericgj/sas2sqlite.git",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={
        "console_scripts": [
            "sas2sqlite=sas2sqlite3.__main__:main",
        ],
    },
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={"test": tests_require},  # for pip
    zip_safe=False,  # for mypy
)



__version__ = "0.0.2"

import sqlite3
from typing import Any, Optional, Iterable, Tuple, List

from sas7bdat import SAS7BDAT  # type: ignore


def import_dataset(
    conn: sqlite3.Connection,
    dataset: SAS7BDAT,
    *,
    table: Optional[str] = None,
    create_table: bool = False,
    sqlite_time_type: str = "TEXT",
    sqlite_date_type: str = "TEXT",
    sqlite_datetime_type: str = "TEXT",
):
    saved_skip_header = dataset.skip_header
    dataset.skip_header = True  # always skip header row for our purposes

    try:
        with dataset as r:
            cols = r.columns
            if table is None:
                table = r.header.properties.name.decode("utf-8")
            with conn:
                if create_table:
                    conn.execute(drop_table_sql(table))
                    conn.execute(
                        create_table_sql(
                            table,
                            cols,
                            encoding=r.encoding,
                            sas_time_formats=r.TIME_FORMAT_STRINGS,
                            sas_date_formats=r.DATE_FORMAT_STRINGS,
                            sas_datetime_formats=r.DATE_TIME_FORMAT_STRINGS,
                            sqlite_time_type=sqlite_time_type,
                            sqlite_date_type=sqlite_date_type,
                            sqlite_datetime_type=sqlite_datetime_type,
                        )
                    )

            with conn:
                for row in r:
                    conn.execute(
                        *insert_sql(
                            table,
                            cols,
                            row,
                            encoding=r.encoding,
                        )
                    )
    finally:
        dataset.skip_header = saved_skip_header


def drop_table_sql(table: str) -> str:
    return f"DROP TABLE IF EXISTS `{table}`;"


def create_table_sql(
    table: str,
    cols: Iterable[Any],
    *,
    encoding: str,
    sas_time_formats: Iterable[str],
    sas_date_formats: Iterable[str],
    sas_datetime_formats: Iterable[str],
    sqlite_time_type: str,
    sqlite_date_type: str,
    sqlite_datetime_type: str,
) -> str:
    col_lines: List[str] = []
    for col in cols:
        if col.type.lower() == "string":
            col_lines.append(f"`{col.name.decode(encoding)}` VARCHAR({col.length})")
        elif col.type.lower() == "number":
            format = None if col.format is None else col.format.upper()
            if format is not None and format in sas_time_formats:
                col_lines.append(f"`{col.name.decode(encoding)}` {sqlite_time_type}")
            elif format is not None and format in sas_date_formats:
                col_lines.append(f"`{col.name.decode(encoding)}` {sqlite_date_type}")
            elif format is not None and format in sas_datetime_formats:
                col_lines.append(
                    f"`{col.name.decode(encoding)}` {sqlite_datetime_type}"
                )
            else:
                col_lines.append(f"`{col.name.decode(encoding)}` NUMERIC")
        else:
            raise ValueError(
                f"Unknown column type '{col.type}': column {col.name.decode(encoding)}"
            )

    col_defs = ", ".join(col_lines)
    return f"CREATE TABLE `{table}` ({col_defs});"


def insert_sql(
    table: str,
    cols: Iterable[Any],
    row: Iterable[Any],
    *,
    encoding: str,
) -> Tuple[str, Iterable[Any]]:
    col_expr = ",".join([f"`{col.name.decode(encoding)}`" for col in cols])
    val_expr = ",".join(["?" for col in cols])
    return (f"INSERT INTO `{table}` ({col_expr}) VALUES ({val_expr});", row)
